"""Module for modelling chemical species."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from massaction.model import ChemModel
from numpy import ndarray

from massaction.constraint import Constraint, ConstraintSweep, ConstraintLike
from massaction.reaction import Reaction

from typing import Union


class Species:
    """Class for modeling chemical species."""

    def __init__(self, species_id: int, parent_model: ChemModel) -> None:
        """Instantiate Species object.

        :param species_id: Species ID number.
        :param parent_model: Parent ChemModel object.
        """
        self.species_id = species_id
        self.parent_model = parent_model
        self.ln_concentration = 0.0

    def __neg__(self) -> FactorSpecies:
        """Return FactorSpecies object representing the negation of a Species object."""
        return FactorSpecies(self, -1.0)

    def __mul__(self, factor: float) -> FactorSpecies:
        """Return FactorSpecies object representing the product of a species and a numerical factor."""
        return FactorSpecies(self, float(factor))

    def __rmul__(self, factor: float) -> FactorSpecies:
        """Return FactorSpecies object representing the product of a species and a numerical factor."""
        return FactorSpecies(self, float(factor))

    def __add__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the sum of two Species objects."""
        return ensure_lincomb(self) + ensure_lincomb(other)

    def __sub__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the difference of two Species objects."""
        return ensure_lincomb(self) - ensure_lincomb(other)

    def __eq__(self, value: Union[float, list[float], ndarray]) -> ConstraintLike:
        """Return Constraint object representing the equality of a Species object and a LinCombSpecies object."""
        if isinstance(value, list) or isinstance(value, ndarray):
            return ConstraintSweep(self, value)
        return Constraint(self, value)

    def __rshift__(self, other: SpeciesLike) -> Reaction:
        """Return Reaction object representing a chemical reaction."""
        return Reaction(self, other)


class FactorSpecies:
    """Class for modeling the union of a chemical species and a numerical pre-factor."""

    def __init__(self, species: Species, factor: float) -> None:
        """Instantiate FactorSpecies object.

        :param species: Species object.
        :param pre_factor: Numerical pre-factor.
        """
        self.species = species
        self.factor = factor

    @property
    def species_id(self) -> int:
        """Return species ID number."""
        return self.species.species_id

    def __neg__(self) -> FactorSpecies:
        """Return FactorSpecies object representing the negation of a FactorSpecies object."""
        return FactorSpecies(self.species, -1.0 * self.factor)

    def __mul__(self, factor: float) -> FactorSpecies:
        """Return FactorSpecies object representing the product of a FactorSpecies and a numerical factor."""
        return FactorSpecies(self.species, self.factor * factor)

    def __rmul__(self, factor: float) -> FactorSpecies:
        """Return FactorSpecies object representing the product of a FactorSpecies and a numerical factor."""
        return FactorSpecies(self.species, self.factor * factor)

    def __add__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the sum of two FactorSpecies objects."""
        return ensure_lincomb(self) + ensure_lincomb(other)

    def __sub__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the difference of two FactorSpecies objects."""
        return ensure_lincomb(self) - ensure_lincomb(other)

    def __eq__(self, value: Union[float, list[float], ndarray]) -> ConstraintLike:
        """Return Constraint object representing the equality of a FactorSpecies object and a numerical value."""
        if isinstance(value, list) or isinstance(value, ndarray):
            return ConstraintSweep(self, value)
        return Constraint(self, value)

    def __rshift__(self, other: SpeciesLike) -> Reaction:
        """Return Reaction object representing a chemical reaction."""
        return Reaction(self, other)


class LinCombSpecies:
    """Class for modeling the linear combination of chemical species."""

    def __init__(self, factor_species_list: list) -> None:
        """Instantiate LinCombSpecies object.

        :param factor_species_list: List of FactorSpecies objects.
        """
        self.factor_species_list = factor_species_list

    def __neg__(self) -> LinCombSpecies:
        """Return LinCombSpecies object representing the negation of a FactorSpecies object."""
        _factor_species_list = [-fs for fs in self.factor_species_list]
        return LinCombSpecies(_factor_species_list)

    def __add__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the sum of two LinCombSpecies objects."""
        other = ensure_lincomb(other)
        return LinCombSpecies([*self.factor_species_list, *other.factor_species_list])

    def __sub__(self, other: SpeciesLike) -> LinCombSpecies:
        """Return LinCombSpecies object representing the difference of two LinCombSpecies objects."""
        return self + (-ensure_lincomb(other))

    def __eq__(self, value: Union[float, list[float], ndarray]) -> ConstraintLike:
        """Return Constraint object representing the equality of a LinCombSpecies object and a numerical value."""
        if isinstance(value, list) or isinstance(value, ndarray):
            return ConstraintSweep(self, value)
        return Constraint(self, value)

    def __rshift__(self, other: SpeciesLike) -> Reaction:
        """Return Reaction object representing a chemical reaction."""
        return Reaction(self, other)

    def print(self, return_str: bool = False) -> str:
        """Print the specifics of the LinCombSpecies object."""
        mystr = ""
        for fs in self.factor_species_list:
            if fs.factor > 0:
                mystr += "+"
            mystr += f"{fs.factor}*species_{fs.species_id} "
        if return_str:
            return mystr
        print(mystr)  # noqa: T201
        return ""


def ensure_lincomb(obj: SpeciesLike) -> LinCombSpecies:
    """Auxiliary function to ensure that a LinCombSpecies object is being used."""
    if isinstance(obj, Species):
        return LinCombSpecies([FactorSpecies(obj, 1.0)])
    if isinstance(obj, FactorSpecies):
        return LinCombSpecies([obj])
    return obj


# Type aliases
SpeciesLike = Union[Species, FactorSpecies, LinCombSpecies]



from massaction.model import ChemModel
from massaction.species import FactorSpecies, LinCombSpecies

from unittest import TestCase

import pytest


class TestSpecies(TestCase):
    """Test the SpeciesLike classes Species, FactorSpecies, and LinCombSpecies."""

    def setUp(self):
        self.model3 = ChemModel(3)
        self.h2o, self.h2, self.o2 = self.model3.get_all_species()

    def test_factor_species(self):
        def fs_assertions(fs, factor, species_obj):
            assert isinstance(fs, FactorSpecies)
            assert fs.factor == factor
            assert fs.species is species_obj

        h2o, h2, o2 = self.h2o, self.h2, self.o2
        # test Species.__rmul__
        fs_assertions(2.0 * h2o, 2.0, h2o)
        # test Species.__mul__
        fs_assertions(h2o * 2.0, 2.0, h2o)
        # test Species.__neg__
        fs_assertions(-o2, -1.0, o2)
        # test FactorSpecies.__neg__
        fs_assertions(-(6.0 * h2), -6.0, h2)
        # test FactorSpecies.__rmul__
        fs_assertions(4.0 * (-3 * h2o), -12.0, h2o)
        # test FactorSpecies.__mul__
        fs_assertions((-3 * h2o) * 4.0, -12.0, h2o)

    def test_lincomb(self):
        def lc_assertions(lc, factors_list, species_list):
            assert isinstance(lc, LinCombSpecies)
            assert len(factors_list) == len(
                species_list
            )  # just internal consistency check
            assert len(lc.factor_species_list) == len(factors_list)
            lc_factors = [fs.factor for fs in lc.factor_species_list]
            lc_species = [fs.species for fs in lc.factor_species_list]
            assert lc_factors == factors_list
            for species1, species2 in zip(lc_species, species_list):
                assert species1 is species2

        h2o, h2, o2 = self.h2o, self.h2, self.o2

        # test Species.__add__
        lc_assertions(h2o + o2, [1.0, 1.0], [h2o, o2])
        # test Species.__sub__
        lc_assertions(h2 - h2o, [1.0, -1.0], [h2, h2o])
        # test FactorSpecies.__add__
        lc_assertions(1.3 * h2o + o2, [1.3, 1.0], [h2o, o2])
        # test FactorSpecies.__sub__
        lc_assertions(1.3 * h2o - 7.0 * o2, [1.3, -7.0], [h2o, o2])
        lc_assertions(1.3 * h2o - (-3.1) * o2, [1.3, 3.1], [h2o, o2])
        # test LinComb.__add__
        lc_assertions((-3.0 * h2 + o2) + 0.6 * h2o, [-3.0, 1.0, 0.6], [h2, o2, h2o])
        # test LinComb.__sub__
        lc_assertions(
            (4.0 * h2o - 11.0 * o2) - 21 * h2, [4.0, -11.0, -21.0], [h2o, o2, h2]
        )

    @pytest.fixture(autouse=True)
    def capsys(self, capsys):
        self.capsys = capsys

    def test_print(self):
        h2o, h2, o2 = self.h2o, self.h2, self.o2
        (h2 + h2o + o2).print()
        captured_text = self.capsys.readouterr().out
        assert captured_text == "+1.0*species_1 +1.0*species_0 +1.0*species_2 \n"

        (-3.0 * h2 + o2).print()
        captured_text = self.capsys.readouterr().out
        assert captured_text == "-3.0*species_1 +1.0*species_2 \n"

        (4.0 * h2o - 11.0 * o2).print()
        captured_text = self.capsys.readouterr().out
        assert captured_text == "+4.0*species_0 -11.0*species_2 \n"

        (-0.1 * h2o - 4.0 * o2 + 3.0 * h2).print()
        captured_text = self.capsys.readouterr().out
        assert captured_text == "-0.1*species_0 -4.0*species_2 +3.0*species_1 \n"



"""Module for the definition of chemical reactions."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from massaction.model import SpeciesLike

import numpy as np


class Reaction:
    """Class for modeling chemical reactions.

    :param reactants: LinCombSpecies object representing the reactants.
    :param products: LinCombSpecies object representing the products.
    """

    def __init__(self, reactants: SpeciesLike, products: SpeciesLike) -> None:
        """Instantiate Reaction object."""
        from massaction.species import ensure_lincomb

        self.reactants = ensure_lincomb(reactants)
        self.products = ensure_lincomb(products)
        self.lincomb = self.products - self.reactants

    def eval(self, ln_concentrations: np.ndarray, ln_eqconst: float) -> float:
        """Evaluate the (logarithmic) mass-action law of the reaction.

        :param ln_concentrations: Array of logarithmic concentrations.
        :param ln_eqconst: logarithmic equilibrium constant.
        """
        factors = np.array([fs.factor for fs in self.lincomb.factor_species_list])
        species_ids = [fs.species_id for fs in self.lincomb.factor_species_list]
        return np.sum(factors * ln_concentrations[species_ids]) - ln_eqconst

    def print(self) -> None:
        """Print the specifics of the Reaction object."""
        reactants_str = self.reactants.print(return_str=True)
        products_str = self.products.print(return_str=True)
        print(f"{reactants_str} >> {products_str}")  # noqa: T201



from massaction.model import ChemModel, get_num_sweep
from massaction.species import Species

from unittest import TestCase

import pytest
import numpy as np


class TestChemModel(TestCase):
    """Test the ChemModel class."""

    def setUp(self):
        self.model1 = ChemModel(1)
        self.model2 = ChemModel(2)
        self.model3 = ChemModel(3)

    def test_get_all_species(self):
        """Test the get_all_species method."""
        list_species = self.model1.get_all_species()
        assert isinstance(list_species, list)
        assert len(list_species) == 1
        assert all(isinstance(s, Species) for s in list_species)
        list_species = self.model2.get_all_species()
        assert isinstance(list_species, list)
        assert len(list_species) == 2
        assert all(isinstance(s, Species) for s in list_species)
        list_species = self.model3.get_all_species()
        assert isinstance(list_species, list)
        assert len(list_species) == 3
        assert all(isinstance(s, Species) for s in list_species)

    def test_get_num_sweep(self):
        h2o, h2, o2 = self.model3.get_all_species()
        c1 = h2o == 1.0
        c2 = h2 == 4.0
        c3 = o2 == [1.0, 2.0, 3.0, 4.0, 5.0]
        assert get_num_sweep([c1, c2, c3]) == 5
        c2 = h2 == [10.0, 11.0, 12.0]
        with pytest.raises(ValueError):
            get_num_sweep([c1, c2, c3])
        c2 = h2 == [10.0, 11.0, 12.0, 13.0, 14.0]
        assert get_num_sweep([c1, c2, c3]) == 5

    def test_solve(self):
        h2o, h2, o2 = self.model3.get_all_species()
        reaction = h2 + o2 >> 2 * h2o
        ln_eqconst = 3.0
        constraint1 = h2o + h2 == 1.0
        constraint2 = h2o + 2 * o2 == 7.0
        # test number of equations
        with pytest.raises(ValueError):
            self.model3.solve([reaction], [ln_eqconst], [constraint1])
        # test number of equilibrium constants
        with pytest.raises(ValueError):
            self.model3.solve(
                [reaction], [ln_eqconst, ln_eqconst], [constraint1, constraint2]
            )
        # test solution for regular constraints
        lnc = self.model3.solve([reaction], [ln_eqconst], [constraint1, constraint2])
        assert np.isclose(2 * lnc[0] - lnc[1] - lnc[2], ln_eqconst)
        c = np.exp(lnc)
        assert np.isclose(c[0] + c[1], 1.0)
        assert np.isclose(c[0] + 2 * c[2], 7.0)
        # test solution for sweep constraints
        constraint2 = o2 == [7.0, 8.0, 9.0]
        lnc_list = self.model3.solve(
            [reaction], [ln_eqconst], [constraint1, constraint2]
        )
        assert len(lnc_list) == 3
        for lnc, val_i in zip(lnc_list, [7.0, 8.0, 9.0]):
            assert np.isclose(2 * lnc[0] - lnc[1] - lnc[2], ln_eqconst)
            c = np.exp(lnc)
            assert np.isclose(c[0] + c[1], 1.0)
            assert np.isclose(c[2], val_i)



from massaction.model import ChemModel

from unittest import TestCase
import numpy as np

import pytest


class TestReaction(TestCase):
    """Test the Reaction class."""

    def setUp(self):
        self.model3 = ChemModel(3)
        h2o, h2, o2 = self.model3.get_all_species()
        self.h2o, self.h2, self.o2 = h2o, h2, o2
        self.reaction = h2 + o2 >> 2 * h2o

    def test_reaction_setups(self):
        def reaction_assertions(
            reaction, reactants, reactants_factors, products, products_factors
        ):
            lnc = np.array([-0.3, 1.5, 2.7])
            lnK = 4.2
            mal_eval = -lnK
            for i, reactant in enumerate(reactants):
                fs = reaction.reactants.factor_species_list[i]
                assert fs.species is reactant
                assert fs.factor == reactants_factors[i]
                mal_eval -= reactants_factors[i] * lnc[reactant.species_id]
            for i, product in enumerate(products):
                fs = reaction.products.factor_species_list[i]
                assert fs.species is product
                assert fs.factor == products_factors[i]
                mal_eval += products_factors[i] * lnc[product.species_id]
            assert np.isclose(reaction.eval(lnc, lnK), mal_eval)

        h2o, h2, o2 = self.h2o, self.h2, self.o2
        # test Species.__rshift__
        reaction = h2 >> 1.9 * o2
        reaction_assertions(reaction, [h2], [1.0], [o2], [1.9])
        # test FactorSpecies.__rshift__
        reaction = 0.4 * h2 >> o2
        reaction_assertions(reaction, [h2], [0.4], [o2], [1.0])
        # test LinCombSpecies.__rshift__
        reaction = (h2 + o2) >> 2 * h2o
        reaction_assertions(reaction, [h2, o2], [1.0, 1.0], [h2o], [2.0])

    @pytest.fixture(autouse=True)
    def capsys(self, capsys):
        self.capsys = capsys

    def test_print_reaction(self):
        reaction = self.reaction
        reaction.print()
        captured_text = self.capsys.readouterr().out
        assert captured_text == "+1.0*species_1 +1.0*species_2  >> +2.0*species_0 \n"



"""Module for defining mass-action laws with constraints."""

from __future__ import annotations


import numpy as np
from scipy.optimize import root

from massaction.constraint import Constraint, ConstraintSweep, ConstraintLike
from massaction.reaction import Reaction
from massaction.species import Species


class ChemModel:
    """Class for modeling chemical species. Manages species and reactions."""

    def __init__(self, num_species: int) -> None:
        """Instantiate ChemModel object."""
        self.num_species = num_species
        self.list_of_species = [Species(i, self) for i in range(num_species)]

    def get_all_species(self) -> list:
        """Return list of all species."""
        return self.list_of_species

    def solve(
        self,
        reactions: list[Reaction],
        ln_eqconst: list[float],
        constraints: list[Constraint],
    ) -> np.ndarray:
        """Solve the system of reactions and constraints.

        :param reactions: List of Reaction objects.
        :param ln_eqconst: Array of logarithmic equilibrium constants.
        :param constraints: List of Constraint objects.
        """
        num_reactions = len(reactions)
        num_constraints = len(constraints)
        num_species = self.num_species
        num_eq = num_reactions + num_constraints
        if num_eq != num_species:
            msg = "Number of equations must be equal to the number of species."
            raise ValueError(msg)
        if len(ln_eqconst) != num_reactions:
            msg = "Number of equilibrium constants must be equal to the number of reactions."
            raise ValueError(msg)

        def eval_system_of_equations(ln_concentrations: np.ndarray) -> np.ndarray:
            ln_equation = np.zeros(num_eq)
            for i in range(num_reactions):
                ln_equation[i] = reactions[i].eval(ln_concentrations, ln_eqconst[i])
            for i in range(num_constraints):
                ln_equation[num_reactions + i] = constraints[i].eval(ln_concentrations)
            return ln_equation

        num_sweep = get_num_sweep(constraints)
        if num_sweep == -1:
            first_guess = np.zeros(num_species)
            result = root(eval_system_of_equations, x0=first_guess)
            return result.x

        results_list = []
        for i in range(num_sweep):
            [cstr.set_current_id(i) for cstr in constraints]
            first_guess = np.zeros(num_species)
            result = root(eval_system_of_equations, x0=first_guess)
            results_list += [result.x]
        return results_list


def get_num_sweep(constraints: list[ConstraintLike]) -> int:
    """Return the number of values in the parameter sweep. If inconsistent, raise error."""
    num_sweep = -1
    for cstr in constraints:
        if not isinstance(cstr, ConstraintSweep):
            continue
        if num_sweep == cstr.num_values:
            continue
        if num_sweep == -1:
            num_sweep = cstr.num_values
            continue
        msg = "All constraint sweeps must have equal number of values"
        raise ValueError(msg)
    return num_sweep



from massaction.model import ChemModel
from massaction.constraint import Constraint, ConstraintSweep
from massaction.species import ensure_lincomb

from unittest import TestCase
import numpy as np

import pytest


class TestConstraint(TestCase):
    """Test the Constraint class."""

    def setUp(self):
        self.model3 = ChemModel(3)

    def test_constraint(self):
        def cstr_assertions(cstr, lincomb, value):
            lnc = np.array([4.7, -6.2, 7.3])
            c = np.exp(lnc)
            cstr_eval_pos = 0.0
            cstr_eval_neg = 0.0
            assert isinstance(cstr, Constraint)
            for fs1, fs2 in zip(
                cstr.lincomb.factor_species_list, lincomb.factor_species_list
            ):
                assert fs1.factor == fs2.factor
                assert fs1.species is fs2.species
                if fs1.factor > 0:
                    cstr_eval_pos += fs1.factor * c[fs1.species.species_id]
                else:
                    cstr_eval_neg -= fs1.factor * c[fs1.species.species_id]
            if value > 0:
                cstr_eval_neg += value
            else:
                cstr_eval_pos -= value
            assert np.isclose(
                cstr.eval(lnc),
                np.log(cstr_eval_pos) - np.log(cstr_eval_neg),
            )
            assert cstr.value == value
            assert cstr.num_values == 1  # by definition, since it is not a sweep

        h2o, h2, o2 = self.model3.get_all_species()

        # test LinCombSpecies.__eq__
        cstr_assertions(
            (3.4 * h2o - 0.2 * h2 - o2) == -1.3,
            ensure_lincomb(3.4 * h2o - 0.2 * h2 - o2),
            -1.3,
        )
        # test FactorSpecies.__eq__
        cstr_assertions(2 * h2o == 10.7, ensure_lincomb(2 * h2o), 10.7)
        # test Species.__eq__
        cstr_assertions(h2 == 4.5, ensure_lincomb(h2), 4.5)

    @pytest.fixture(autouse=True)
    def capsys(self, capsys):
        self.capsys = capsys

    def test_print(self):
        h2o, h2, o2 = self.model3.get_all_species()
        # test Species.__eq__
        (h2 == 4.5).print()
        out, _ = self.capsys.readouterr()
        assert out == "+1.0*species_1  == 4.5\n"
        # test FactorSpecies.__eq__
        (2 * h2o == 10.7).print()
        out, _ = self.capsys.readouterr()
        assert out == "+2.0*species_0  == 10.7\n"
        # test LinCombSpecies.__eq__
        (h2 - h2o == -1.3).print()
        out, _ = self.capsys.readouterr()
        assert out == "+1.0*species_1 -1.0*species_0  == -1.3\n"


class TestConstraintSweep(TestCase):
    """Test the ConstraintSweep class."""

    def setUp(self):
        self.model3 = ChemModel(3)
        self.h2o, self.h2, self.o2 = self.model3.get_all_species()

    def test_constraint_sweep(self):
        def cs_assertions(cs, lincomb, values):
            assert isinstance(cs, ConstraintSweep)
            for fs1, fs2 in zip(
                cs.lincomb.factor_species_list, lincomb.factor_species_list
            ):
                assert fs1.factor == fs2.factor
                assert fs1.species is fs2.species
            assert cs.num_values == len(values)
            assert cs.current_id == 0
            assert cs.value == values[0]
            for i in range(len(values)):
                cs.set_current_id(i)
                assert cs.current_id == i
                assert cs.values[i] == values[i]
                assert cs.value == values[i]
            # test cs.eval ?

        h2o, h2, o2 = self.h2o, self.h2, self.o2
        # test Species.__eq__
        cs_assertions(h2o == [1.0, 2.0, 3.0], ensure_lincomb(h2o), [1.0, 2.0, 3.0])
        cs_assertions(
            o2 == np.linspace(0.4, 0.7, 5), ensure_lincomb(o2), np.linspace(0.4, 0.7, 5)
        )
        # test FactorSpecies.__eq__
        cs_assertions(
            7 * h2o == [6.0, 1.0, 8.0], ensure_lincomb(7 * h2o), [6.0, 1.0, 8.0]
        )
        cs_assertions(
            2 * o2 == np.linspace(-2.0, 9, 3),
            ensure_lincomb(2 * o2),
            np.linspace(-2.0, 9, 3),
        )
        # test LinCombSpecies.__eq__
        cs_assertions(h2o - h2 + o2 == [2.0, 5.0, 4.0], h2o - h2 + o2, [2.0, 5.0, 4.0])
        cs_assertions(
            3 * o2 - h2 == np.linspace(12.0, -11.0, 4),
            3 * o2 - h2,
            np.linspace(12, -11.0, 4),
        )



import setuptools

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="massaction",
    python_requires="~=3.8",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Module for the definition of chemical constraints."""

from __future__ import annotations

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from massaction.model import SpeciesLike

import numpy as np


class Constraint:
    """Class for modeling constraints on chemical species."""

    def __init__(self, lincomb: SpeciesLike, value: float) -> None:
        """Instantiate Constraint object.

        :param lincomb: LinCombSpecies object.
        :param value: Value of the constraint.
        """
        from massaction.species import ensure_lincomb

        self.lincomb = ensure_lincomb(lincomb)
        self.value = value

    def eval(self, ln_concentrations: np.ndarray) -> float:
        """Evaluate the constraint.

        :param ln_concentrations: Array of logarithmic concentrations.
        """
        concentrations = np.exp(ln_concentrations)
        factors = np.array([fs.factor for fs in self.lincomb.factor_species_list])
        species_ids = [fs.species_id for fs in self.lincomb.factor_species_list]
        mask_pos = factors > 0.0
        mask_neg = factors < 0.0
        val_pos = (
            np.sum(factors[mask_pos] * concentrations[species_ids][mask_pos])
            if mask_pos.any()
            else 0.0
        )
        val_neg = (
            -np.sum(factors[mask_neg] * concentrations[species_ids][mask_neg])
            if mask_neg.any()
            else 0.0
        )
        if self.value > 0:
            return np.log(val_pos) - np.log(val_neg + self.value)
        return np.log(val_pos + (-self.value)) - np.log(val_neg)

    def print(self) -> None:
        """Print the specifics of the Constraint object."""
        mystr = self.lincomb.print(return_str=True)
        print(f"{mystr} == {self.value}")  # noqa: T201

    @property
    def num_values(self) -> int:
        """Return the number of values in the parameter sweep (regular constraint: 1)."""
        return 1

    def set_current_id(self, current_id: int) -> None:
        """Set new id of value in list of parameter sweep values (irrelevant in regular constraint).

        :param current_id: the new id
        """
        pass


class ConstraintSweep(Constraint):
    """Class to represent a parameter sweep in a constraint."""

    def __init__(self, lincomb: SpeciesLike, values: list | np.ndarray):
        self.values = values
        self.current_id = 0
        super().__init__(lincomb, self.values[self.current_id])

    @property
    def num_values(self) -> int:
        """Return the number of values in the parameter sweep."""
        return len(self.values)

    def set_current_id(self, current_id: int) -> None:
        """Set new id of value in list of parameter sweep values.

        :param current_id: the new id
        """
        self.current_id = current_id
        self.value = self.values[self.current_id]


ConstraintLike = Union[Constraint, ConstraintSweep]
